/*
   QUICKB.C - Quick B Protocol Support routines

 	originally converted to C by Paul M. Resch
	heavly modified by John Bridges
 */

/*

  This source was originally derived from QUICKB.INC, version 121687, written
  by Russ Ranshaw, CompuServe Incorporated.

 */

#define	TRUE	1
#define	FALSE	0

#define	DLE		16
#define	ETX		03
#define	NAK		21
#define	ENQ		05
#define	CR		0x0D
#define	LF		0x0A
#define	MAXBUFSIZE	1032		/* Largest data block we can handle */
#define	MAXSA		2		/* Maximum number of waiting packets */

#define	DEFBUFSIZE	511		/* Default data block */
#define	DEFWS		1		/* I can send 2 packets ahead */
#define	DEFWR		1		/* I can receive single send-ahead */
#define	DEFBS		8		/* I can handle 1024 bytes */
#define	DEFCM		1		/* I can handle CRC */
#define	DEFDQ		1		/* I can handle non-quoted NUL */

#define	MAXERRORS	10


/* Receive states */

#define	RGETDLE	0
#define	RGETB		1
#define	RGETSEQ	2
#define	RGETDATA	3
#define	RGETCHECKSUM	4
#define	RSENDACK	5
#define	RTIMEDOUT	6
#define	RSUCCESS	7

/* Send states */

#define	SGETDLE	1
#define	SGETNUM	2
#define	SHAVEACK	3
#define	SGETPACKET	4
#define	STIMEDOUT	5
#define	SSENDNAK	6
#define	SSENDENQ	7
#define	SSENDDATA	8

typedef	struct	PACKETB {
	int seq;	/* Packet's sequence number */
	int num;	/* Number of bytes in packet */
	unsigned char buf[MAXBUFSIZE]; /* Actual packet data */
	} PACKET;

static PACKET	SABuf[MAXSA+1];  /* Send-ahead buffers */

/* Table of control characters that need to be masked */

static char masktable[]={
		0, 0, 0, 1, 0, 1, 0, 0,	/* NUL SOH SOB ETX EOT ENQ SYN BEL */
		0, 0, 0, 0, 0, 0, 0, 0,	/* BS  HT  LF  VT  FF  CR  SO  SI */
		1, 1, 0, 1, 0, 1, 0, 0,	/* DLE DC1 DC2 DC3 DC4 NAK ^V  ^W */
		0, 0, 0, 0, 0, 0, 0, 0};  /* CAN ^Y  ^Z  ESC ?	?	?	? */

static char hexdigit[]="0123456789ABCDEF";

unsigned int checksum=0;		/* May hold CRC */
static int seqnum;		/* Current Sequence Number - init by TermENQ */
static int rsize;		/* size of receiver buffer */
static unsigned int scounter,rcounter;
static int timedout;		/* we timed out before receiving character */
static int cchar;		/* current character */
static int masked;		/* TRUE if ctrl character was 'masked' */
static int packetreceived;	/* True if a packet was received */
static unsigned char rbuffer[MAXBUFSIZE];

	/* Other End's Parameters */
static char HisWS;		/* Sender's Window Send */
static char HisWR;		/* Sender's Window Receive */
static char HisBS;		/* Sender's Block Size */
static char HisCM;		/* Sender's Check Method */

	/* Negotiated Parameters */
static char OurWS;		/* Negotiated Window Send */
static char OurWR;		/* Negotiated Window Receive */
static char OurBS;		/* Negotiated Block Size */
static char OurCM;		/* Negotiated Check Method */

static int QuickB;		/* True if Quick B in effect */
static int UseCRC;		/* True if CRC in effect */
static int buffersize;		/* OurBS * 4 */
static int SAMax;		/* 1 if SA not enabled, else MAXSA */
static int SAEnabled;		/* True if Send-Ahead is permitted */
static int ackSA;		/* Which SABuf is waiting for an ACK */
static int fillSA;		/* Which SABuf is ready for new data */
static int SAWaiting;		/* Number of SABuf's waiting for ACK */
static int Aborting;		/* TRUE if aborting transfer */

extern unsigned long lseek(int fs,long pos,int mode);


extern int dosio;
extern int curx,cury;
extern unsigned char attrib;

#define cur(yy,xx)  (curx=xx,cury=yy)

#define tval (*(unsigned long far *)0x0040006cl)

static unsigned long oldt;
static unsigned long oldrt;
static unsigned long filelen;
static unsigned long opos;
static unsigned int bknum;

#define sposx (35)
#define sposy (5)
#define statx (45)
#define staty (12)

static unsigned int savr[statx][staty];
static int savx,savy;
static unsigned char sava=0;
static int errors,toterrs;

begstat(fd,fname)
int fd;
char *fname;
{
	oldt=tval;
	bknum=1;
	errors=0;
	toterrs=0;
	filelen=lseek(fd,0l,2);
	lseek(fd,0l,0);
	opos=0;
	if(dosio)
	{
		sput("Transmitting \"");
		sput(fname);
		sput("\"\r\n\r\n");
	}
	else
	{
		savx=curx;
		savy=cury;
		sava=attrib;
		attrib=((attrib&15)<<4)|(attrib>>4);
		tbox(sposx,sposy,statx,staty,savr);
		cur(sposy,sposx+1);
		putchar('[');
		sput(fname);
		putchar(']');
		cur(sposy+2,sposx+1);
		sput(" Transmit time     :\r\n");
		if(filelen)
		{
			sput(" Time remaining    :\r\n");
			sput(" Bytes to transmit :\r\n");
		}
		else
			sput("\n\n");

		sput(" Block             :\r\n");
		sput(" Bytes transmited  :\r\n\n");
		sput(" Current errors    :     Average CPS :\r\n");
		sput(" Total errors      :     Current CPS :\r\n");
		upderr();
	}
}

endstat(fd)
int fd;
{
	if(fd) close(fd);
	if(!dosio && sava)
	{
		unbox(sposx,sposy,statx,staty,savr);
		curx=savx;
		cury=savy;
		attrib=sava;
		sava=0;
	}
}

statbk()
{
	oldrt=tval;
	if(dosio)
	{
		sput("Block #");
		putint(bknum,4);
	}
	else
	{
		cur(sposy+5,sposx+21);
		putint(bknum,4);
	}
	++bknum;
}

status(fd)
int fd;
{
	unsigned int cps,rcps,rt;
	unsigned long pos;

	pos=lseek(fd,0l,1);
	if(tval<oldt)
		oldt-=1573040l;
	cps=(unsigned int)(((pos*4661l)/(tval-oldt))>>8);
	if(tval<oldrt)
		oldrt-=1573040l;
	rcps=(unsigned int)((((pos-opos)*4661l)/(tval-oldrt))>>8);
	rt=(filelen-pos)/cps;
	if(dosio)
	{
		sput(" Bytes:");
		putlint(pos,9);
		sput(" avgCPS:");
		putint(cps,4);
		sput(" rawCPS:");
		putint(rcps,4);
		if(filelen)
		{
			sput("  ");
			puttime(rt);
			sput(" Time Remain");
		}
		putchar('\r');
	}
	else
	{
		if(filelen)
		{
			cur(sposy+3,sposx+21);
			puttime(rt);
			cur(sposy+4,sposx+21);
			putlint(filelen-pos,9);
		}
		cur(sposy+6,sposx+21);
		putlint(pos,9);
		cur(sposy+8,sposx+21+18);
		putint(cps,4);
		cur(sposy+9,sposx+21+18);
		putint(rcps,4);
	}
	opos=pos;
}

upderr()
{
	if(!dosio && sava)
	{
		cur(sposy+2,sposx+21);
		puttime((unsigned int)(((tval-oldt)<<8)/4661l));
		cur(sposy+8,sposx+21);
		putint(errors,3);
		cur(sposy+9,sposx+21);
		putint(toterrs,3);
	}
}

/* Update the checksum/CRC */

static void dochecksum(chr)
unsigned char chr;
{
	if(QuickB && UseCRC)
		checksum=updCRC(chr);
	else
		checksum=updCHK(chr);
}

static void sendfailure(code)
int code;
{
	register PACKET	*p;

	ackSA=0;
	fillSA=0;
	SAWaiting=0;
	Aborting=TRUE;		/* inform getACK we're aborting */

	p=&SABuf[0];
	p->buf[0]='F';
	p->buf[1]=code;

	if(sendpacket(1))
		SAFlush();	/* Gotta wait for the host to ACK it */
}

/*
 * bpENQ is called when the terminal emulator receives the character <ENQ>
 * from the host.  Its purpose is to initialize for B Protocol and tell the
 * host that we support Quick B.
 */

void bpENQ()
{
	seqnum=0;
	buffersize=DEFBUFSIZE;	/* Set up defaults */
	QuickB=FALSE;			/* Not Quick B Protocol */
	UseCRC=FALSE;			/* Not CRC16 */
	SAEnabled=FALSE;		/* No Send-Ahead by us */
	SAMax=1;			/* single packet sent */

	cputc(DLE);
	cputc('+');

	cputc(DLE);
	cputc('0');
}

static void sendmaskedbyte(c)
int c;
{
	c&=0xff;
	if(c<32)
	{
		if(masktable[c]!=0)
		{
			cputc(DLE);
			cputc(c+'@');
		}
		else
			cputc(c);
	}
	else
		cputc(c);

	scounter=(scounter+1)&DEFBUFSIZE;
}

static void sendack()
{
	cputc(DLE);
	cputc(seqnum + '0');
}

static void sendnak()
{
	cputc(NAK);
}


static void sendenq()
{
	cputc(ENQ);
}

static int readbyte()
{
	timedout=FALSE;

	cchar=cgetc(100);

	if(cchar<0)
		return(FALSE);

	rcounter=(rcounter+1)&DEFBUFSIZE;
	return(TRUE);
}


static int readmaskedbyte()
{
	masked=FALSE;

	if(readbyte()==FALSE)
		return(FALSE);

	if(cchar==DLE)
	{
		if(readbyte()==FALSE)
			return(FALSE);
		cchar &=0x1F;
		masked=TRUE;
	}

	return(TRUE);
}

/* Increment Sequence Number */

static int incrseq(value)
int value;
{
	return(value==9 ? 0 : value + 1);
}

/*	LeadinSeen is TRUE if the <DLE><B> has been seen already.
	fromsendpacket is TRUE if called from SendPacket
	(causes exit on first error detected)
	Returns True if packet is available from host. */

static int readpacket(leadinseen,fromsendpacket)
int leadinseen,fromsendpacket;
{
register int state;
	int nextseq,blocknum,newcks;
register int i;

	packetreceived=FALSE;
	for(i=0; i<buffersize; i++)
		rbuffer[i]=0;
	nextseq=incrseq(seqnum);
	errors=0;
	upderr();

	if(leadinseen)		/* Start off on the correct foot */
		state=RGETSEQ;
	else
		state=RGETDLE;

	while(TRUE)
	{
		switch (state){
		case RGETDLE :
			if(userabort())
			{
				sendfailure('A');
				return(FALSE);
			}

			if(!readbyte())
					state=RTIMEDOUT;
			else if((cchar & 0x7F)==DLE)
					state=RGETB;
			else if((cchar & 0x7F)==ENQ)
				state=RSENDACK;
			break;

		case RGETB :
			if(!readbyte())
				state=RTIMEDOUT;
			else if((cchar & 0x7F)=='B')
				state=RGETSEQ;
			else if(cchar==ENQ)
				state=RSENDACK;
			else
				state=RGETDLE;
			break;

		case RGETSEQ :
			if(!readbyte())
				state=RTIMEDOUT;
			else if(cchar==ENQ)
				state=RSENDACK;
			else
			{
				if(QuickB && UseCRC)
					checksum=initCRC(-1);
				else 
					checksum=initCRC(0);

				blocknum=cchar - '0';

				dochecksum(cchar);

				i=0;
				state=RGETDATA;
			}
			break;

		case RGETDATA :
			rcounter=0;
			if(!readmaskedbyte())
				state=RTIMEDOUT;
			else if((cchar==ETX)&& !masked)
			{
				dochecksum(ETX);
				state=RGETCHECKSUM;
			}
			else
			{
				rbuffer[i]=cchar;
				i=i+1;
				dochecksum(cchar);
			}
			break;

		case RGETCHECKSUM :
			if(!readmaskedbyte())
				state=RTIMEDOUT;
			else
			{
				if(QuickB && UseCRC)
				{
					checksum=updCRC(cchar);

					if(!readmaskedbyte())
						newcks=checksum ^ 0xFF;
					else
					{
						checksum=updCRC(cchar);
						newcks=0;
					}
				}
				else newcks=cchar;

				if(newcks !=checksum)
					state=RTIMEDOUT;
				else if(rbuffer[0]=='F')/* Watch for Failure Packet */
					state=RSUCCESS;	/* which is accepted regardless */
				else if(blocknum==seqnum)	/* Watch for duplicate block */
					state=RSENDACK;
				else if(blocknum !=nextseq)
					state=RTIMEDOUT;	/* Bad sequence number */
				else
					state=RSUCCESS;
			}
			break;

		case RTIMEDOUT :
			++errors;
			++toterrs;
			upderr();
			if((errors>MAXERRORS)||(fromsendpacket))
				return(FALSE);

			sendnak();

			if(fromsendpacket)
				return(FALSE);

			state=RGETDLE;
			break;

		case RSENDACK :
			sendack();
			state=RGETDLE;	/* wait for the next block */
			break;

		case RSUCCESS :
			seqnum=blocknum;
			rsize=i;
			packetreceived=TRUE;
			return(TRUE);
		}
	}
}

static void senddata(BufferNumber)
int BufferNumber;
{
	register int i;
	register PACKET	*p;

	scounter=0;
	p=&SABuf[BufferNumber];
	if(QuickB && UseCRC)
		checksum=initCRC(-1);
	else 
		checksum=initCRC(0);

	cputc(DLE);
	cputc('B');

	cputc(p->seq + '0');
	dochecksum(p->seq + '0');

	for(i=0; i<=p->num; i++)
	{
		sendmaskedbyte(p->buf[i]);
		dochecksum(p->buf[i]);
	}

	cputc(ETX);
	dochecksum(ETX);

	if(QuickB && UseCRC)
		sendmaskedbyte(checksum >> 8);

	sendmaskedbyte(checksum);
}

static int incrSA(oldvalue)
int oldvalue;
{
	return(oldvalue==MAXSA ? 0 : oldvalue + 1);
}

/*
 * ReSync is called to restablish syncronism with the remote.  This is
 * accomplished by sending <ENQ><ENQ> and waiting for the sequence
 * <DLE><d><DLE><d> to be received, ignoring everything else.
 *
 * Return is -1 on time out, else the digit <d>.
 */
#define	GETFIRSTDLE		1
#define	GETFIRSTDIGIT		2
#define	GETSECONDDLE		3
#define	GETSECONDDIGIT	4

static int ReSync()
{
register int state,Digit1;

	cputc(ENQ);	/* Send <ENQ><ENQ> */
	cputc(ENQ);
	state=GETFIRSTDLE;

	while(1)
	{
		switch(state){
		case GETFIRSTDLE :
			if(!readbyte())
				return(-1);
			if(cchar==DLE)
				state=GETFIRSTDIGIT;
			break;
		case GETFIRSTDIGIT :
			if(!readbyte())
				return(-1);
			if((cchar >='0')&&(cchar <='9'))
			{
				Digit1=cchar;
				state=GETSECONDDLE;
			}
			break;
		case GETSECONDDLE :
			if(!readbyte())
				return(-1);
			if(cchar==DLE)
				state=GETSECONDDIGIT;
			break;
		case GETSECONDDIGIT :
			if(!readbyte())
				return(-1);
			if((cchar >='0')&&(cchar <='9'))
			{
				if(Digit1==cchar)
					return(cchar);
				else
				{
					Digit1=cchar;
					state=GETSECONDDLE;
				}
			}
			else
				state=GETSECONDDLE;
			break;
		}
	}
}

/*
 * getACK is called to wait until the SABuf indicated by ackSA
 * has been ACKed by the host.
 */
static int getACK()
{
	register int state,i;
	int blocknum;
	int SentENQ;
	int SAIndex;

	packetreceived=FALSE;
	errors=0;
	upderr();
	SentENQ=FALSE;
	state=SGETDLE;

	while(TRUE)
	{
		switch(state){
		case SGETDLE :
			if(userabort())
			{
				sendfailure('A');
				return(FALSE);
			}

			if(!readbyte())
				state=STIMEDOUT;
			else if(cchar==DLE)
				state=SGETNUM;
			else if(cchar==NAK)
			{
				++errors;
				++toterrs;
				upderr();
				if(errors > MAXERRORS)
					return(FALSE);
				state=SSENDENQ;
			}
			else if(cchar==ETX)
				state=SSENDNAK;
			break;

		case SGETNUM :
			if(!readbyte())
				state=STIMEDOUT;
			else if((cchar >='0')&&(cchar <='9'))
				state=SHAVEACK;	/* Received ACK */
			else if(cchar=='B')
				state=SGETPACKET; /* Try to get packet */
			else if(cchar==NAK)
			{
				++errors;
				++toterrs;
				upderr();
				if(errors > MAXERRORS)
					return(FALSE);
				state=SSENDENQ;
			}
			else
				state=STIMEDOUT;
			break;

		case SGETPACKET :
			if(readpacket(TRUE,TRUE))
			{
				if(rbuffer[0]=='F')
				{
					sendack();
					return(FALSE);
				}
				else
					return(TRUE);
			}

			state=STIMEDOUT; /* On a bad receive, try again */
			break;
		case SHAVEACK:
			blocknum=cchar - '0';
			if(SABuf[ackSA].seq==blocknum)
			{	/* This is the one we're waiting for */
				ackSA=incrSA(ackSA);
				SAWaiting--;
				return(TRUE);
			}
			else if(SABuf[incrSA(ackSA)].seq==blocknum &&
				(SAWaiting==2))
			{	/* Must have missed an ACK */
				ackSA=incrSA(ackSA);
				ackSA=incrSA(ackSA);
				SAWaiting -=2;
				return(TRUE);
			}
			else if(SABuf[ackSA].seq==incrseq(blocknum))
			{
				if(SentENQ)
					state=SSENDDATA;
				else
					state=SGETDLE;
			}
			else
				state=Aborting ? SGETDLE : STIMEDOUT;
			SentENQ=FALSE;
			break;
		case STIMEDOUT :
			++errors;
			++toterrs;
			upderr();
			if(errors > MAXERRORS)
				return(FALSE);
			else
			{
				if(Aborting && (errors > 3))
					return(FALSE);
			}
			state=SSENDENQ;
			break;

		case SSENDNAK :
			++errors;
			++toterrs;
			upderr();
			if(errors > MAXERRORS)
				return(FALSE);

			sendnak();

			state=SGETDLE;
			break;

		case SSENDENQ :
			++errors;
			++toterrs;
			upderr();
			if(errors > MAXERRORS)
				return(FALSE);
			cchar=ReSync();
			if(cchar==-1)
				state=SSENDENQ;
			else
				state=SHAVEACK;
			SentENQ=TRUE;
			break;

		case SSENDDATA :
			SAIndex=ackSA;

			for(i=1; i<=SAWaiting; i++)
			{
				senddata(SAIndex);
				SAIndex=incrSA(SAIndex);
			}

			state=SGETDLE;
			SentENQ=FALSE;
			break;
		}
	}
} /* getACK */

static int sendpacket(size)
int size;
{
	if(SAWaiting==SAMax)
		if(!getACK())
			return(FALSE);

	seqnum=incrseq(seqnum);
	SABuf[fillSA].seq=seqnum;
	SABuf[fillSA].num=size;
	senddata(fillSA);
	fillSA=incrSA(fillSA);
	SAWaiting=SAWaiting + 1;
	return(TRUE);
}
/*
 * SAFlush is called after sending the last packet to get host's
 * ACKs on outstanding packets.
 */
static int SAFlush()
{
	while(SAWaiting > 0)
		if(!getACK())
			return(FALSE);
	return(TRUE);
}

/* SendFile is called to send a file to the host */
static int sendfile(name)
char name[];
{
	int fd,n;
	register PACKET	*p;

	fd=open(name);
	if(fd<=0)
  	{
		sput("\r\n** Cannot find file...aborting **\r\n");
		sendfailure('E');
		return(FALSE);
	}

	filelen=lseek(fd,0l,2);
	lseek(fd,0l,0);

	begstat(fd,name);
	do
	{
		p=&SABuf[fillSA];
		p->buf[0]='N';
		n=read(fd,&p->buf[1],buffersize);
		if(n>0)
		{
			statbk();
			if(sendpacket(n)==FALSE)
			{
				endstat(fd);
				return(FALSE);
			}
			status(fd);
		}
	} while(n==buffersize);

	endstat(fd);

	if(n < 0)
	{
		sendfailure('E');
		sput("\r\n** Read failure...aborting **\r\n");
		return(FALSE);
	}

/* Inform host that the file was sent */
	p=&SABuf[fillSA];
	p->buf[0]='T';
	p->buf[1]='C';

	if(sendpacket(2)==FALSE)
		return(FALSE);
	else
	{
		sput("\r\nWaiting for host...\r\n");
		if(!SAFlush())
			return(FALSE);
		return(TRUE);
	}
}

/*
 * dotransportparameters is called when a Packet type of + is received.
 * It sends a packet of our local Quick B parameters and sets the Ourxx
 * parameters to the minimum of the sender's and our own parameters.
 */
static dotransportparameters()
{
	register PACKET	*p;

	HisWS=rbuffer[1];	/* Pick out Sender's parameters */
	HisWR=rbuffer[2];
	HisBS=rbuffer[3];
	HisCM=rbuffer[4];

	p=&SABuf[fillSA];
	p->buf[0]='+';  /* Prepare to return our own parameters */
	p->buf[1]=DEFWS;
	p->buf[2]=DEFWR;
	p->buf[3]=DEFBS;
	p->buf[4]=DEFCM;
	p->buf[5]=DEFDQ;

	if(!sendpacket(5))
		return;

	if(SAFlush())		/* Wait for host's ACK on our packet */
	{
/* Take minimal subset of Transport Params. */
/* If he can send ahead, we can receive it. */
		OurWR=(HisWS < DEFWR)? HisWS : DEFWR;

/* If he can receive send ahead, we can send it. */
		OurWS=(HisWR < DEFWS)? HisWR : DEFWS;

		OurBS=HisBS < DEFBS ? HisBS : DEFBS;

		OurCM=HisCM < DEFCM ? HisCM : DEFCM;

		if(OurBS==0)
			OurBS=4;	/* Default */

		buffersize=OurBS * 128;

		QuickB=TRUE;

		if(OurCM==1)
			UseCRC=TRUE;

		if(OurWS !=0)
		{
			SAEnabled=TRUE;
			SAMax=MAXSA;
		}
	}
}

/*
  doapplicationparameters is called when a ? packet is received.
  This version ignores the host's packet and returns a ? packet
  saying that normal B Protocol File Transfer is supported.
 (Well, actually it says that no extended application packets are
	supported.  The T packet is assumed to be standard.)*/

static void doapplicationparameters()
{
	register PACKET	*p;

	p=&SABuf[fillSA];
	p->buf[0]='?';	/* Build the ? packet */
	p->buf[1]=1;		/* The T packet flag */

	if(sendpacket(1))		/* Send the packet */
		SAFlush();
}

/* ReceiveFile is called to receive a file from the host */
int receivefile(name)
char name[];
{
	int fd;

	fd=create(name);
	if(fd<=0)
	{
		sput("\r\n** Cannot create file...aborting **\r\n");
		sendfailure('E');
		return(FALSE);
	}

	sendack();

	begstat(fd,name);

/* Process each incoming packet until 'TC' packet received or failure */
	while(TRUE)
	{
		statbk();
		if(readpacket(FALSE,FALSE))
		{
			switch(rbuffer[0])
			{
			case 'N' :
				if(write(fd,&rbuffer[1],rsize-1)!=rsize-1)
				{
					endstat(fd);
					sput("\r\n** Write failure...aborting **\r\n");
					sendfailure('E');
					return(FALSE);
				}
				sendack();
				status(fd);
				break;

			case 'T' :
				if(rbuffer[1]=='C')
				{
					endstat(fd);
					sendack();
					return(TRUE);
				}
				else
				{
					endstat(fd);
					sput("\r\n** Invalid termination packet...aborting **\r\n");
					sendfailure('N');
					return(FALSE);
				}

			case 'F' :
				endstat(fd);
				sendack();
				sput("\r\n** Failure packet received...aborting **\r\n");
				return(FALSE);
			}
		}
		else
		{
			endstat(fd);
			sput("\r\n** Failed to receive packet...aborting **\r\n");
			return(FALSE);
		}
	}
}

/*
 * bpDLE is called from the main program when the character <DLE> is
 * received from the host.
 *
 * This routine calls readpacket and dispatches to the appropriate
 * handler for the incoming packet.
 */
void bpDLE()
{
register int i;
	char filename[255];
register char *pt;
/*
 * Begin by getting the next character.  If it is <B> then enter the
 * BProtocol state.  Otherwise simply return.
 */

	if(cgetc(10)!='B')
		return;

	ackSA=0;	/* Initialize Send-ahead variables */
	fillSA=0;
	SAWaiting=0;
	Aborting=FALSE;	/* not aborting ... yet */
/*  <DLE><B> received; begin B Protocol */

	rcounter=0;
	scounter=0;

	if(dosio)
	{
		sput("\r\n");
		if(QuickB)
		{
			sput("*** Quick B is in effect ***\r\n");

			if(UseCRC)
				sput("*** Using CRC ***\r\n");

			if(OurWS!=0)	/* Allow send-ahead if other end agrees */
				sput("*** Send-Ahead enabled ***\r\n");
		}
	}

	if(readpacket(TRUE,FALSE))
	{
/* Dispatch on the type of packet just received */

		switch(rbuffer[0])
		{
		case 'T':	/* File Transfer Application */
			switch(rbuffer[1])
			{
			case 'D' :	/* downloading */
				break;
			case 'U' :	/* uploading */
				break;
			default :
				sendfailure('N');
				return;
			}

			switch(rbuffer[2])
			{
			case 'A':	/* ascii file */
				break;
			case 'B':	/* binary file */
				break;
			default :
				sendfailure('N');	/* not implemented */
				return;
			}

			i=2;
			pt=filename;
			while((rbuffer[i]!=0)&&(i<rsize-1))
				*pt++=rbuffer[++i];

			*pt='\0';

			if(rbuffer[1]=='U')
				i=sendfile(filename);
			else
				i=receivefile(filename);

			if(dosio)
			{
				sput("\r\n");
				if(i)
					sput("\r\nTransfer successfully completed!\r\n");
			}
			break;

		case '+':	/* Received Transport Parameters Packet */
			dotransportparameters();
			break;

		case '?':	/* Received Application Parameters Packet */
			doapplicationparameters();
			break;

		default:	/* Unknown packet; tell host we don't know */
			sendfailure('N');
				break;

		}  /* of case */
	}	/* of if readpacket the */
}


